import pygame
import random


# barvy pro kostičky tetrisu

colors = [
    (0, 0, 0),
    (255, 0, 0),
    (0, 255, 0),
    (0, 0, 255)
]


class Figure:
    x = 0
    y = 0

    # tvary kostiček a jejich rotace, 0 = 0°, 1 = 90°, 2 = 180°, 3 = 270°
    figures = [
        [[1, 5, 9, 13], [4, 5, 6, 7], [1, 5, 9, 13], [4, 5, 6, 7]],
        [[4, 5, 9, 10], [2, 6, 5, 9], [4, 5, 9, 10], [2, 6, 5, 9]],
        [[6, 7, 9, 10], [1, 5, 6, 10], [6, 7, 9, 10], [1, 5, 6, 10]],
        [[1, 2, 5, 9], [0, 4, 5, 6], [1, 5, 9, 8], [4, 5, 6, 10]],
        [[1, 2, 6, 10], [5, 6, 7, 9], [2, 6, 10, 11], [3, 5, 6, 7]],
        [[1, 4, 5, 6], [1, 4, 5, 9], [4, 5, 6, 9], [1, 5, 6, 9]],
        [[1, 2, 5, 6], [1, 2, 5, 6], [1, 2, 5, 6], [1, 2, 5, 6]],
    ]

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.type = random.randint(0, len(self.figures) - 1)  # náhodný výběr tvaru kostičky
        self.color = random.randint(1, len(colors) - 1)  # výběr barvy pro tuto kostičku
        self.rotation = 0  # výchozí rotace bude s indexem 0

    def image(self): # vrácení tvaru objektu s požadovanou rotací
        return self.figures[self.type][self.rotation]

    def rotate(self):  # otáčení objektu
        self.rotation = (self.rotation + 1) % len(self.figures[self.type])


class Tetris:
    level = 2  # obtížnost - vyšší číslo zvyšuje rychlost padání kostiček
    score = 0  # skore, počítá se ze zničených plných řádků
    state = "start"  # stav hry, poté z něho zjišťujeme, jestli jsme prohráli nebo stále hrajeme
    field = []  # hrací pole - bude obsahovat již položené kostičky tetrisu
    height = 0  # výchozí výška herního plánu
    width = 0  # výchozí šířka herního plánu
    x = 100  # umístění hracího plánu v okně na souřadnici X
    y = 60  # umístění hracího plánu v okně na souřadnici Y
    zoom = 20  # škálování velikosti hry
    figure = None

    def __init__(self, height, width):
        self.height = height
        self.width = width
        self.field = []
        self.score = 0
        self.state = "start"
        for i in range(height):
            new_line = []
            for j in range(width):
                new_line.append(0)
            self.field.append(new_line)

    def new_figure(self):  #vytvoření nové kostičky
        self.figure = Figure(3, 0)

    def intersects(self):  # metoda zjišťující, jestli kostička přesahuje výšku hracího plánu, nebo pokud se dotýká jiné kostičky či hrany, pokud ano, vráti true
        intersection = False
        for i in range(4):
            for j in range(4):
                if i * 4 + j in self.figure.image():
                    if i + self.figure.y > self.height - 1 or \
                            j + self.figure.x > self.width - 1 or \
                            j + self.figure.x < 0 or \
                            self.field[i + self.figure.y][j + self.figure.x] > 0:
                        intersection = True
        return intersection

    def break_lines(self):  # Zničení řádku kostiček na herním plánu, když je celý řádek obsazen
        lines = 0
        for i in range(1, self.height):
            zeros = 0
            for j in range(self.width):
                if self.field[i][j] == 0:
                    zeros += 1
            if zeros == 0:
                lines += 1
                for i1 in range(i, 1, -1):
                    for j in range(self.width):
                        self.field[i1][j] = self.field[i1 - 1][j]
        self.score += lines ** 2

    def go_space(self): # Metoda pro okamžité položení kostičky na nejnížší možnou souřadnici Y
        while not self.intersects():
            self.figure.y += 1
        self.figure.y -= 1
        self.freeze()

    def go_down(self): # posun kostričky o jedno pole dolů
        self.figure.y += 1
        if self.intersects():
            self.figure.y -= 1
            self.freeze()

    def freeze(self):  # položení kostičky a zapamatování její pozice, přidání do herního plánu
        for i in range(4):
            for j in range(4):
                if i * 4 + j in self.figure.image():
                    self.field[i + self.figure.y][j + self.figure.x] = self.figure.color
        self.break_lines()
        self.new_figure()
        if self.intersects():
            self.state = "gameover"

    def go_side(self, dx): # posun kostičky do strany
        old_x = self.figure.x
        self.figure.x += dx
        if self.intersects():
            self.figure.x = old_x

    def rotate(self):  # otočení kostičky
        old_rotation = self.figure.rotation
        self.figure.rotate()
        if self.intersects():
            self.figure.rotation = old_rotation


# spuštění instance pygame
pygame.init()

# základní barvy využité pro okno s hrou
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
GRAY = (128, 128, 128)

size = (400, 500) # výška a šířka našeho pygame okna
screen = pygame.display.set_mode(size)

pygame.display.set_caption("hra tetris") # nastavení názvu okna

done = False # proměnná, dle které budeme posuzovat, jestli se má hra vypnout nebo zůstat otevřená
clock = pygame.time.Clock()
fps = 20 # Dále v kodu tuto proměnnou využíváme pro výpočet rychlosti hry na základě levelu
game = Tetris(20, 10) # vytvoření herního plánu o výšce 20 kostiček, šířce 10 kostiček
counter = 0 # počítadlo ticků hry

pressing_down = False  # Proměnná pro uložení stavu stisknutých kláves

while not done:  #dokud proměnná done je false
    if game.figure is None:
        game.new_figure()
    counter += 1 # přičteme tick
    if counter > 100000:
        counter = 0

    if counter % (fps // game.level // 2) == 0 or pressing_down: # pokud zbytek po dělení operace je 0, případně pokud je stisknutá jakákoliv klávesa, provedeme akci
        if game.state == "start":
            game.go_down()

    for event in pygame.event.get():
        if event.type == pygame.QUIT: # Pokud chceme ukončit aplikaci, nastavíme proměnnou done na true a hra se v dalším ticku vypne.
            done = True
        if event.type == pygame.KEYDOWN: # Pokud je stisknuta jakákoliv klávesa ze seznamu níže, vykonám akci dle stisknuté klávesy
            if event.key == pygame.K_UP:
                game.rotate()
            if event.key == pygame.K_DOWN:
                pressing_down = True
            if event.key == pygame.K_LEFT:
                game.go_side(-1)
            if event.key == pygame.K_RIGHT:
                game.go_side(1)
            if event.key == pygame.K_SPACE:
                game.go_space()
            if event.key == pygame.K_ESCAPE:
                game.__init__(20, 10)

    if event.type == pygame.KEYUP:
            if event.key == pygame.K_DOWN:
                pressing_down = False

    screen.fill(WHITE) # Vyplníme pygame okno bílou barvou

    # vytvoření herního plánu o výšce game.height a šířce game.width
    for i in range(game.height):
        for j in range(game.width):
            pygame.draw.rect(screen, GRAY, [game.x + game.zoom * j, game.y + game.zoom * i, game.zoom, game.zoom], 1)
            if game.field[i][j] > 0:
                pygame.draw.rect(screen, colors[game.field[i][j]],
                                 [game.x + game.zoom * j + 1, game.y + game.zoom * i + 1, game.zoom - 2, game.zoom - 1])

    if game.figure is not None:
        for i in range(4):
            for j in range(4):
                p = i * 4 + j
                if p in game.figure.image():
                    pygame.draw.rect(screen, colors[game.figure.color],
                                     [game.x + game.zoom * (j + game.figure.x) + 1,
                                      game.y + game.zoom * (i + game.figure.y) + 1,
                                      game.zoom - 2, game.zoom - 2])

    font = pygame.font.SysFont('Calibri', 25, True, False)  # vytvoření fontu pro nápisy
    font1 = pygame.font.SysFont('Calibri', 65, True, False) # vytvoření fontu pro nápisy
    text = font.render("Score: " + str(game.score), True, BLACK) # zobrazení aktuálního score černou barvou
    text_game_over = font1.render("Game Over", True, (255, 125, 0))
    text_game_over1 = font1.render("Press ESC", True, (255, 215, 0))

    screen.blit(text, [0, 0]) # zobrazení textu z proměnné text na pozici 0, 0
    if game.state == "gameover": #pokud je hra u konce, zobrazíme texty z proměnných text_game_over a text_game_over1 na pozicích definovaných níže
        screen.blit(text_game_over, [20, 200])
        screen.blit(text_game_over1, [25, 265])

    pygame.display.flip() # obnovíme celý obsah pygame okna
    clock.tick(fps) # zavolání dalšího ticku

pygame.quit() # po ukončení cyklu se uzavře program
